;;
;;                      Copyright (c) 1998 SQUARE USA, Inc.
;;                             All Rights Reserved.
;;
;;  SQUARE USA Inc. hereby grants permission to use, copy, modify, and
;;  distribute this software and its documentation for any purpose, provided
;;  that existing copyright notices are retained verbatim in all copies.
;;
;;  IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
;;  FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
;;  ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
;;  DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
;;  POSSIBILITY OF SUCH DAMAGE.
;;
;;  THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
;;  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
;;  FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
;;  IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
;;  NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
;;  MODIFICATIONS.
;;
;;  $Id: cgitool.scm,v 1.1 1999/11/10 11:08:20 shiro Exp $
;;  

(require "html-classes")
(provide "cgitool")

;;;
;;; Useful procedures to write CGI script in Scheme.
;;;

;; [References]
;;   For HTTP header specification, see RFC2068.
;;   For Cookie handling, see RFC2109.
;;   For syntax of QUERY_STRING, see HTML/4.0 specification.

;;================================================================
;; HTTP Header and Cookie handling
;;

;;-----------------------------------------------------------------
;; cgi:write-header &key :content-type :cookies :location
;;   Write out CGI header string.  Keyword argument :content-type
;;   specifies Content-Type header, which defaulted "text/html".
;;
;;   Keyword argument :location specifies location header of http,
;;   to redirect user's request to new location.
;;   This option supercedes content-type option.
;;
;;   Keyword argument :cookies specifies "cookies" to be sent to
;;   the client.   It is an assoc-list and each element must be
;;   the following form:
;;       (NAME VALUE [keyword value ...])
;;
;;   Keyword-value pairs are intended for per-cookie attribute values as
;;   defined in RFC2109 (new specification).  It also supports "old"
;;   cookie format in original Netscape proposal
;;   (http://www.netscape.com/newsref/std/cookie_spec.html).
;;   To make two standard coexist, VALUEs are not quoted.
;;
;;   Following attribute specification (keyword-value pair) are recognized.
;;   If the attribute is only available either new spec or old spec, it is
;;   marked as "(new)" or "(old)".  If you mix new-spec-only attribute and
;;   old-spec-only attribute, the result will be unknown.  Note that if you
;;   want to use new specification, you must specify "Version" attribute
;;   explicitly (it should be 1).
;;
;;      :comment STRING    Comment attribute.  (new)
;;      :domain STRING     Domain attribute.
;;      :max-age INTEGER   Max-Age attribute.  (new)
;;      :path STRING       Path attribute.
;;      :secure BOOLEAN    Secure attribute.
;;      :version INTEGER   Version attribute.  (new)
;;      :expires DATE      Expires attribute.  (old)
;;         If DATE is integer, it must be a seconds from
;;         Jan 1, 1970 0:0:0 GMT.  If DATE is a string, it must be
;;         a formatted date string ("Wdy, DD-Mon-YY HH:MM:SS GMT").

(define (cgi:write-header . opts)
  (let ((content-type (get-keyword :content-type opts "text/html"))
        (cookies      (get-keyword :cookies opts '()))
        (location     (get-keyword :location opts #f)))
    (if location
        (format #t "Location: ~a\n" location)
        (format #t "Content-type: ~a\n" content-type))
    (map cgi:write-cookie-string cookies)
    (newline)  ;; terminates RFC822 header section
    ))

(define (cgi:write-cookie-string spec)
  (format #t "Set-cookie: ~a=~a" (car spec) (cadr spec))
  (let loop ((keys (cddr spec)))
    (if (or (null? keys) (null? (cdr keys)))
        (newline)
        (let* ((key  (car keys))
               (knam (keyword->string key))
               (ksym (substring knam 1 (string-length knam)))
               (value (cadr keys)))
          (case key
            ((:comment :domain :max-age :path :version)
             (format #t ";~a=~a" ksym value))
            ((:secure)
             (if value (format #t ";~a" ksym) #f))
            ((:expires)
             (if (integer? value)
                 (format #t ";Expires=~a"
                         (posix-strftime "%a, %d-%b-%y %T GMT"
                                         (posix-gmtime value)))
                 (format #t ";Expires=~a" value)))
            (else
             (error "CGI:make-cookie-string: Unrecognized keyword for Cookie: ~s" spec)))
          (loop (cddr keys)))
        )))

;;----------------------------------------------------------------
;; cgi:parse-cookie &optional COOKIE
;;   Parse encoded cookie and returns assoc-list.  If COOKIE is 
;;   omitted, environment variable HTTP_COOKIE is used.
;; 
;;   The "new" format as of RFC2109 can be handled, although all
;;   additional attributes (attributes starts with "$") will be
;;   ignored.

(define (cgi:parse-cookie . rest)
  (let* ((cookie (if (pair? rest) (car rest)
                     (or (getenv "HTTP_COOKIE") "")))
         (len    (string-length cookie))
         (e      (lambda ()
                   (cgi:error "Badly formed cookie string : ~s" cookie)))
         (tokens (or (cgi:lex-cookie cookie len 0 0 'none) (e)))
         (next-token (lambda () (if (null? tokens)
                                    #f
                                    (let ((r (car tokens)))
                                      (set! tokens (cdr tokens))
                                      r)))))
    (let loop ((t (next-token)))
      (cond ((not t) '())
            ((string? t)
             (let ((s (string->symbol t))
                   (n (next-token)))
               (cond ((not n) `((,t)))
                     ((memq n '(#\;)) `((,s) ,@(loop (next-token))))
                     ((eq? n #\=)
                      (let ((v (next-token)))
                        (cond ((not v) `((,s "")))
                              ((string? v)
                               (let ((d (next-token)))
                                 (if (or (not d) (memq d '(#\; #\,)))
                                     `((,s ,v) ,@(loop (next-token)))
                                     (e))))
                              (else (e)))))
                     (else (e)))))
            (else (e))))
    ))

(define (cgi:lex-cookie str len head ptr state)
  (if (>= ptr len)
      (case state
        ((quoted) #f)                   ;error
        ((token) `(,(substring str head ptr)))
        (else '()))
      (let ((c     (string-ref str ptr))
            (ptr+1 (+ ptr 1))
            (head+1 (+ head 1)))
        (case c
          ((#\space #\newline #\tab)
           (case state
             ((quoted) (cgi:lex-cookie str len head ptr+1 state))
             ((token)  `(,(substring str head ptr)
                         ,@(cgi:lex-cookie str len ptr+1 ptr+1 'none)))
             (else     (cgi:lex-cookie str len head+1 ptr+1 state))))
          ((#\")
           (case state
             ((quoted) `(,(substring str head+1 ptr)
                         ,@(cgi:lex-cookie str len ptr+1 ptr+1 'none)))
             ((token)  `(,(substring str head ptr)
                         ,@(cgi:lex-cookie str len ptr ptr+1 'quoted)))
             (else     (cgi:lex-cookie str len ptr ptr+1 'quoted))))
          ((#\= #\; #\,)
           (case state
             ((quoted) (cgi:lex-cookie str len head ptr+1 state))
             ((token)  `(,(substring str head ptr)
                         ,c
                         ,@(cgi:lex-cookie str len ptr+1 ptr+1 'none)))
             (else     `(,c
                         ,@(cgi:lex-cookie str len ptr+1 ptr+1 'none)))))
          (else
           (case state
             ((none)   (cgi:lex-cookie str len ptr ptr+1 'token))
             (else     (cgi:lex-cookie str len head ptr+1 state)))))
        )
      )
  )
 
;;================================================================
;; Error Page
;; 

;;-----------------------------------------------------------------
;; *cgi:error-page-html-options*
;;   A list of init-keywords and values, which will be used to
;;   create <html-page> object for error page.
;;   Example: '(:page-title "CGI script erro" :author-name "shiro")

(define *cgi:error-page-html-options* #f)

;;-----------------------------------------------------------------
;; cgi:error-page FORMAT . ARGS
;;    Generate error page (<html-page>) object

(define (cgi:error-page fmt . args)
  (let* ((text (make <html-text>
                     :string (apply format #f fmt args)))
         (page (apply make <html-page>
                      :contents (list text)
                      (or *cgi:error-page-html-options*
                          '()))))
    page))

;;-----------------------------------------------------------------
;; cgi:error FORMAT . ARGS
;;    Print out error page and exit

(define (cgi:error fmt . args)
  (let ((page (apply cgi:error-page fmt args)))
    (cgi:write-header)
    (write-html page)
    (exit 0))
  )

;;================================================================
;; Parsing Query String
;; 

;;-----------------------------------------------------------------
;; cgi:parse-query-string &optional STRING
;;    Parse query string STRING and return alist of result.
;;    When STRING is omitted, the value of environment variable QUERY_STRING
;;    is used if it is set, or just return '() if it is not set.
;;    If provided string is illegal for quoted query string, this 
;;    procedure generates a html error page and exits.

(define (cgi:parse-query-string . rest)
  (let* ((string (if (pair? rest)
                     (car rest)
                     (or (getenv "QUERY_STRING") "")))
         (tokens (cgi:parse-query-string-lex string (string-length string)
                                             0 '()))
         (next-token
          (lambda () (if (null? tokens) #f 
                         (let ((t (car tokens)))
                           (set! tokens (cdr tokens)) t))))
         (result '())
         (add-entry
          (lambda (name . value)
            (let ((p (assq name result)))
              (if p
                  (if (pair? value) (set-cdr! p (cons (car value) (cdr p))))
                  (set! result (cons (cons name value) result))))))
         (e (lambda () (cgi:error "Malformed query string: ~s" string)))
         )
    (let loop ((token (next-token)))
      (cond ((not token) result)
            ((string? token)
             (let* ((name (string->symbol token))
                    (entry (assq name result))
                    (next (next-token)))
               (case next
                 ((#\& #f) (add-entry name) (loop (next-token)))
                 ((#\=) (let ((value (next-token)))
                          (if (not value) (e)
                              (let ((next (next-token)))
                                (if (not (memq next '(#\& #f))) (e)
                                    (begin (add-entry name value)
                                           (loop (next-token))))))))
                 (else (e)))))
            (else (e))))
    )
  )

(define (cgi:parse-query-string-lex str len ptr r)
  (if (>= ptr len) `(,(list->string (reverse r)))
      (let ((c     (string-ref str ptr))
            (ptr+1 (+ ptr 1)))
        (case c
          ((#\+) (cgi:parse-query-string-lex str len ptr+1 `(#\space ,@r)))
          ((#\%) (if (> ptr (- len 2))
                     (cgi:error "Bad query string (wrong % syntax): ~s" str)
                     (let* ((ptr+3 (+ ptr 3))
                            (c (string->number (substring str ptr+1 ptr+3)
                                               16)))
                       (if (not c)
                           (cgi:error "Bad query string (wrong % syntax): ~s"
                                      str))
                       (cgi:parse-query-string-lex str len ptr+3
                                                   `(,(integer->char c)
                                                     ,@r)))))
          ((#\= #\&)
           `(,(list->string (reverse r))
             ,c
             ,@(cgi:parse-query-string-lex str len ptr+1 '())))
          (else (cgi:parse-query-string-lex str len ptr+1 `(,c ,@r)))))))

;; for backward compatibility

(define cgi:url-unquote cgi:parse-query-string)

;;-----------------------------------------------------------------
;; cgi:get-parameter-value NAME PLIST &key :default :error :multivalue :convert
;;    PLIST is a list of parameters which cgi:parse-query-string or
;;    cgi:parse-cookie returns, and NAME is a name of parameter
;;    what you want.
;;    If NAME is not found in PLIST, (1) if a keyword argument :default
;;    is specified, it is returned, (2) otherwise, if a keyword argument
;;    :error is specified, call it as a procedure with one argument NAME,
;;    (3) otherwise, generates html error page and exits.
;; 
;;    If keyword argument :multivalue is true, the procedure returns
;;    list of value(s).  Otherwise, the procedure returns a string,
;;    concatenating all the value(s) with whitespace inbetween.

(define (cgi:get-parameter-value name param . opts)
  (let* ((unbound "UnBoUnD")
         (default (get-keyword :default opts unbound))
         (eproc   (get-keyword :error opts #f))
         (multi   (get-keyword :multivalue opts #f))
         (conv    (get-keyword :convert opts (lambda (x) x)))
         (bound?  (lambda (x) (not (eq? x unbound)))))
    (cond ((assq name param)
           => (lambda (p)
                (if multi
                    (map conv (cdr p))
                    (conv
                     (let loop ((p (cdr p)))
                       (cond ((null? p) "")
                             ((null? (cdr p)) (car p))
                             (else
                              (string-append (car p) " " (loop (cdr p)))))))))
           )
          ((bound? default) default)
          (eproc (eproc name))
          (else (cgi:error "Requied query parameter \"~s\" is not provided."
                           name)))
    )
  )

;;================================================================
;; High-level Macro
;; 

;;-----------------------------------------------------------------
;; cgi:let ARGUMENT-SPECS . BODY
;;   This macro is ultimately convenient way to write a CGI script.
;;   It retrieves a query-string parameter and cookie values from
;;   environment variable QUERY_STRING and HTTP_COOKIE, respectively,
;;   and bound it to local variables according to ARGUMENT-SPECS,
;;   then executes BODY.
;;   ARGUMENT-SPECS must be a list, in which each element is one of
;;   the followings:
;;
;;     SYMBOL : look for parameter whose name is SYMBOL from query
;;              parameters, and bind its value to SYMBOL.  If named
;;              parameter is not found, cgi:error is called.
;;              The special case are the symbol &queries and &cookies.
;;              The symbol &queries is bound to the alist which
;;              cgi:parse-query-string returns, and &cookies is bound
;;              to the alist which cgi:parse-cookie returns.
;;
;;     (SYMBOL . OPTIONS) : OPTIONS are used to retrieve parameter
;;              value.  SYMBOL is bound to the value.
;;              Possible options are:
;; 
;;              :name NAME
;;                 Specifies parameter name other than SYMBOL.
;;              :source SOURCE
;;                 SOURCE might be either :query or :cookie.  If it is
;;                 :cookie, the value is retrieved from cookie instead
;;                 of query string.  The default is :query.
;;
;;              :convert PROC
;;              :default DEFAULT-VALUE
;;              :error ERROR-PROC
;;              :multivalue BOOLEAN
;;                 Given to cgi:get-parameter-value.
;;

(define-macro (cgi:let specs . body)
  (let* ((params  (gensym))
         (cookies (gensym))
         (getsrc  (gensym))
         (get-args   (lambda (spec)
                       (let* ((opts (if (symbol? spec) '() (cdr spec)))
                              (unbound "UnBoUnD")
                              (default (get-keyword :default opts unbound))
                              (eproc   (get-keyword :error opts unbound))
                              (mvalue  (get-keyword :multivalue opts unbound))
                              (convert (get-keyword :convert opts unbound))
                              (bound? (lambda (v) (not (eq? v unbound)))))
                         `(,@(if (bound? default) `(:default ,default) '())
                           ,@(if (bound? eproc)   `(:error ,eproc) '())
                           ,@(if (bound? mvalue)  `(:multivalue ,mvalue) '())
                           ,@(if (bound? convert) `(:convert ,convert) '())
                           ))))
         (get-name   (lambda (spec)
                       (cond ((symbol? spec) `',spec)
                             ((get-keyword :name (cdr spec) #f))
                             (else `',(car spec)))))
         (get-symbol (lambda (spec)
                       (if (symbol? spec) spec (car spec))))
         (get-src    (lambda (spec)
                       (cond ((symbol? spec) :query)
                             ((get-keyword :source (cdr spec) :query))
                             (else :query))))
         )
    `(let* ((,params  (cgi:parse-query-string))
            (,cookies (cgi:parse-cookie))
            (,getsrc  (lambda (which)
                        (if (eq? which :cookie) ,cookies ,params))))
       (let ,(map (lambda (spec)
                    (cond ((memq spec '(&all &queries))
                           `(,spec ,params))
                          ((eq? spec '&cookies)
                           `(,spec ,cookies))
                          (else
                           `(,(get-symbol spec)
                             (cgi:get-parameter-value
                              ,(get-name spec)
                              (,getsrc ,(get-src spec))
                              ,@(get-args spec))))))
                  specs)
         ,@body))
    ))

;; Test code   

;(cgi:let ((name :default "anonymous" :convert string-upper)
;          &queries
;          &cookies)
;  (let ((a (html-paragraph
;            :contents `(,(format #f "Hello, ~a" name))))
;        (b (html-paragraph
;            :contents `(,(format #f "Queries = ~s" &queries))))
;        (c (html-paragraph
;            :contents `(,(format #f "Cookies = ~s" &cookies)))))
;    (cgi:write-header :cookies '((cgitool "$Id: cgitool.scm,v 1.1 1999/11/10 11:08:20 shiro Exp $")))
;    (write-html (html-page :page-title "cgitool test"
;                           :author-name  "shiro"
;                           :author-email "shiro@squareusa.com"
;                           :contents `(,a ,b ,c)))))


