;;;
;;; genmake.scm - generate Makefile
;;;
;;;   Copyright (c) 1999 Shiro Kawai (shiro@acm.org)
;;;   Permission to use, modify, and distribute of this code is granted
;;;   under the same condition as of STk.
;;;
;;; $Id: genmake.scm,v 1.8 1999/11/08 03:12:10 shiro Exp $
;;;

;;; This module is a modification of stk-genmake script to make it
;;; more customizable and extensible.  I borrowed the idea from Perl's
;;; ExtUtils::MakeMaker.
;;;
;;; Each extention is provided with a file called Makefile.scm, which
;;; looks like this:
;;;
;;;  (require "genmake")
;;;  (generate-makefile :target-scm "myextension.scm"
;;;                     :target-so '(("myextension" "myextension" "moresource"))
;;;                     :libs  '("-lmylib")
;;;                     :lpath '("-L/usr/home/my/lib")
;;;   )
;;;
;;; See $(DOC)/genmake.html for details.
;;;

(define-module genmake
  (export genmake-version>=?
          genmake-stkversion>=?
          generate-makefile)

;;; Default

;;; The Large portion of code is taken from stk-genmake by
;;;  Eric Gallesio [eg@unice.fr]
;;;  Copyright (c) 1998-1999 Erick Gallesio - I3S-CNRS/ESSI <eg@unice.fr>

(define genmake-version "0.4")
(define config.make-file	; Complete path name of the  Config.make file 
  (string-append (%library-location) "/" (machine-type) "/Config/config.make"))

(define (abort . l)
  (apply format (current-error-port) l)
  (newline (current-error-port))
  (exit 0))

(define (warn . l)
  (apply format (current-error-port)l)
  (newline (current-error-port)))

;;;
;;; Version checking
;;;

(define (version>=? v1 v2)
  (let loop ((sv1 (map string->number (split-string v1 ".")))
             (sv2 (map string->number (split-string v2 "."))))
    (cond ((null? sv2) #t)
          ((not (car sv2)) #t)
          ((null? sv1) #f)
          ((not (car sv1)) #f)
          ((> (car sv1) (car sv2)) #t)
          ((< (car sv1) (car sv2)) #f)
          (else (loop (cdr sv1) (cdr sv2))))))

;;;
;;; Parse config.make
;;;

(define *parameters* '())               ;alist of parameters
(define *so-dependencies* '())          ;alist of dependency between .so & .o
(define *parameter-hook* #f)
(define *prelude-hook* #f)
(define *postlude-hook* #f)

(define rx-comment  (string->regexp "^[ \t]*#"))
(define rx-continue (string->regexp "\\\\$"))
(define rx-define (string->regexp "^[ \t]*([_A-Za-z][_0-9A-Za-z]*)[ \t]*=[ \t]*(.*)"))

(define (parse-config-make)
  (if (not (file-exists? config.make-file))
      (abort "~A: File ~S does not exist (you probably need a \"make install.libs\""
             *program-name*))
  (with-input-from-file config.make-file
    (lambda ()
      (let loop ((line (read-line)) (pending '()))
        (cond
         ((eof-object? line)
          (if (not (null? pending))
              (warn "warning: unterminated line at the end of config.make" )))
         ((rx-comment line) (loop (read-line) pending))
         ((rx-continue  line)
          => (lambda (match)
               (loop (read-line)
                     (cons (substring line 0 (caar match)) line))))
         (else
          (push-parameter (apply string-append
                                 (reverse (cons line pending))))
          (loop (read-line) '()))))))
  )

(define (push-parameter string)
  (cond ((rx-define string)
         => (lambda (match)
              (let ((var (substring string
                                    (car (cadr match)) (cadr (cadr match))))
                    (val (substring string
                                    (car (caddr match)) (cadr (caddr match)))))
                (set! *parameters* (cons (list var val) *parameters*)))))
        (else
         (warn "warning: unrecognized line in config.make: ~a" string))))

;;;
;;; Configure parameters
;;;

(define (set-parameter param value)
  (let ((a (assoc param *parameters*))
        (v (if (pair? value) value (list value))))
    (if a
        (set-cdr! a v)
        (set! *parameters* (cons (cons param v) *parameters*)))))

(define (add-parameter param value)
  (let ((a (assoc param *parameters*))
        (v (if (pair? value) value (list value))))
    (if a
        (set-cdr! a (append! (cdr a) v))
        (set! *parameters* (cons (cons param v) *parameters*)))))

(define (get-parameter param)
  (cond ((assoc param *parameters*) => (lambda (p) (cdr p)))
        (else #f)))

;; interface to be passed to hook procedures
(define (parameter-interface op name . arg)
  (let ((value (if (null? arg) "" (car arg))))
    (case op
      ((:get) (get-parameter name))
      ((:set) (set-parameter name value))
      ((:add) (add-parameter name value))
      (else (abort "parameter: unknown operation: ~s" op)))))

(define (add-so-dependency sofile ofiles)
  (let ((a (assoc sofile *so-dependencies*)))
    (if a
        (set-cdr! a (append! (cdr a) ofiles))
        (set! *so-dependencies* (cons (cons sofile ofiles) *so-dependencies*)))))

(define (target-defined)
  (assoc "TARGET" *parameters*))

(define (target-scm-defined)
  (assoc "TARGET_SCM" *parameters*))

(define (target-so-defined)
  (assoc "TARGET_SO" *parameters*))

(define (package-defined)
  (assoc "PACKAGE" *parameters*))

(define (package-version)
  (assoc "PACKAGE_VERSION" *parameters*))

(define (test-scm-defined)
  (assoc "TEST_SCM" *parameters*))

;; Mapping from keyword to parameter.

(define *parameter-list*
  ;; keyword   add? parameter
  `((:prefix    #f "prefix")
    (:stkdir    #f "stkdir")
    (:libdir    #f "libdir")
    (:execdir   #f "execdir")
    (:confdir   #f "confdir")
    (:ardir     #f "ardir")
    (:incdir    #f "incdir")
    (:mandir    #f "mandir")
    (:bindir    #f "bindir")
    (:dflgs     #t "DFLGS")
    (:eobj      #t "EOBJ")
    (:etkobj    #t "ETKOBJ")

    (:ipath     #t "IPATH")             ;extra -Isomething
    (:lpath     #t "LPATH")             ;extra -Lsomething
    (:libs      #t "LIBS")              ;extra -lsomething

    (:target-scm #t "TARGET_SCM")
    (:package    #f "PACKAGE")              ;package name (used for tar)
    (:package-version #f "PACKAGE_VERSION") ;package version (used for tar)

    (:test-scm  #t "TEST_SCM")
    ))

(define (process-parameters keylist)
  (let loop ((keylist keylist))
    (cond ((null? keylist) (adjust-target) (adjust-package))
          ((or (null? (cdr keylist))
               (not (keyword? (car keylist))))
           (abort "generate-makefile: Bad keyword argument: ~s" (car keylist)))
          ((assq (car keylist) *parameter-list*)
           => (lambda (entry)
                ((if (cadr entry) add-parameter set-parameter)
                 (caddr entry)
                 (cadr keylist))
                (loop (cddr keylist))))
          ((eq? (car keylist) :target-so)
           (process-target-so (cadr keylist))
           (loop (cddr keylist)))
          ((eq? (car keylist) :prelude-hook)
           (if (not (procedure? (cadr keylist)))
               (abort "generate-makefile: prelude-hook is not a procedure: ~s"
                      (cadr keylist)))
           (set! *prelude-hook* (cadr keylist))
           (loop (cddr keylist)))
          ((eq? (car keylist) :postlude-hook)
           (if (not (procedure? (cadr keylist)))
               (abort "generate-makefile: postlude-hook is not a procedure: ~s"
                      (cadr keylist)))
           (set! *postlude-hook* (cadr keylist))
           (loop (cddr keylist)))
          ((eq? (car keylist) :parameter-hook)
           (if (not (procedure? (cadr keylist)))
               (abort "generate-makefile: parameter-hook is not a procedure: ~s"
                      (cadr keylist)))
           (set! *parameter-hook* (cadr keylist))
           (loop (cddr keylist)))
          (else
           (abort "generate-makefile: Unknown keyword: ~s" (car keylist))))))

(define (add-sh-suffix base)
  (string-append base ".$(SH_SUFFIX)"))

(define (add-o-suffix base)
  (string-append base ".o"))

(define (process-target-so desc)
  (cond ((string? desc)
         (add-parameter "TARGET_SO" (add-sh-suffix desc))
         (add-so-dependency desc (list desc)))
        ((list? desc)
         (for-each (lambda (desc)
                     (cond ((string? desc)
                            (add-parameter "TARGET_SO"
                                           (add-sh-suffix desc))
                            (add-so-dependency desc (list desc)))
                           ((list? desc)
                            (add-parameter "TARGET_SO"
                                           (add-sh-suffix (car desc)))
                            (add-so-dependency (car desc) (cdr desc)))
                           (else
                            (abort "generate-makefile: bad spec in :target-so: ~s" desc))))
                   desc))
        (else (abort "generate-makefile: bad spec in :target-so : ~s" desc))
        ))

(define (adjust-target)
  (if (target-scm-defined)
      (add-parameter "TARGET" "$(TARGET_SCM)"))
  (if (target-so-defined)
      (add-parameter "TARGET" "$(TARGET_SO)")))

;; when a package name isn't specified, use current directory name if possible.
(define (adjust-package)
  (unless (package-defined)
    (let ((cwd (split-string (getcwd) "/")))
      (if (pair? cwd)
          (add-parameter "PACKAGE" (car (last-pair cwd))))))
  (unless (package-version)
    (add-parameter "PACKAGE_VERSION" "0.1"))
  )

;;;
;;; Write Makefile
;;;

(define (write-parameters)
  (for-each (lambda (p)
              (display (car p))
              (let ((len (string-length (car p)))) ;make it look nice
                (if (< len 16)
                    (dotimes (n (- 15 len)) (display #\ ))))
              (display " = ")
              (let loop ((value (cdr p)))
                (cond ((null? value) (newline))
                      ((null? (cdr value)) (display (car value)) (newline))
                      (else (display (car value))
                            (display #\ )
                            (loop (cdr value))))))
            (reverse *parameters*)))

(define (generate-prelude)
  (format #t "# Makefile automatically generated by genmake version ~A. DO NOT EDIT\n"   
	  genmake-version)
  (write-parameters)
  ;; Generate the .SUFFIXES rules
  (format #t "\n")
  (format #t "CFLAGS= $(SH_CCFLAGS) $(STKCFLAGS) $(DFLGS) -DUSE_TK  -DHAVE_UNISTD_H=1 -DHAVE_LIMITS_H=1 -DHAVE_SIGACTION=1 $(IPATH) -I$(incdir)\n")
  (format #t "LDFLAGS = $(SH_LDFLAGS)\n\n")
  (format #t "\nall: $(TARGET)\n\n"))

(define (generate-target)
  (for-each (lambda (dep)
              (format #t "~a.$(SH_SUFFIX):" (car dep))
              (for-each (lambda (ofile)
                          (display " ") (display (add-o-suffix ofile)))
                        (cdr dep))

              (format #t "\n\t$(SH_LOADER) $(LDFLAGS) $@ ")
              (for-each (lambda (ofile)
                          (display " ") (display (add-o-suffix ofile)))
                        (cdr dep))
              (format #t " $(LPATH) $(LIBS)\n")
              (format #t "\tif test -f a.out  ;then mv a.out $*.$(SH_SUFFIX); fi\n\n")

              )
            *so-dependencies*)
  (newline))

(define (generate-postlude)
  (format #t "install: $(TARGET)\n")
  (when (target-scm-defined)
     (format #t "\tfor f in $(TARGET_SCM); do\\\n")
     (format #t "\t  $(CP) -f $$f $(stkdir)/site-scheme/$$f;\\\n")
     (format #t "\t  chmod 444 $(stkdir)/site-scheme/$$f;\\\n")
     (format #t "\tdone\n"))
  (when (target-so-defined)
     (format #t "\tfor f in $(TARGET_SO); do\\\n")
     (format #t "\t  $(CP) -f $(ardir)/$$f;\\\n")
     (format #t "\t  chmod 444 $(ardir)/$$f;\\\n")
     (format #t "\tdone\n"))
  (format #t "\n")
  (format #t "clean:\n")
  (format #t "\t@/bin/rm -f *.o *.$(SH_SUFFIX) core *~\n\n")
  (format #t "realclean: clean\n")
  (format #t "\t@/bin/rm -f Makefile\n\n")
  (when (package-defined)
     (format #t "dist: Manifest\n")
     (format #t "\tcd ..; tar cvf - `cat $(PACKAGE)/Manifest | sed 's/^\\(.*\\)$$/$(PACKAGE)\\/\\1/'` | gzip > $(PACKAGE)-$(PACKAGE_VERSION).tar.gz\n\n"))
  (when (test-scm-defined)
     (format #t "test: $(TARGET) $(TEST_SCM)\n")
     (format #t "\t@for f in $(TEST_SCM); do stk -f $$f; done\n\n"))
  (format #t "#End of Makefile\n"))

;;;
;;; Main entry
;;;

(define (GENERATE-MAKEFILE . args)
  (with-output-to-file "Makefile"
    (lambda ()
      (set! *parameters* '())
      (parse-config-make)
      (process-parameters args)
      (if *parameter-hook* (*parameter-hook* parameter-interface))
      (generate-prelude)
      (if *prelude-hook* (*prelude-hook* parameter-interface))
      (generate-target)
      (generate-postlude)
      (if *postlude-hook* (*postlude-hook* parameter-interface))
      (exit 0))))

(define (GENMAKE-VERSION>=? vers)
  (unless (string? vers)
     (abort "genmake-version>=? : string expected: ~s" vers))
  (unless (version>=? genmake-version vers)
     (abort "genmake: version ~a or higher expected, but your genmake is version ~a.  Please obtain the newer version.\n"
            vers genmake-version))
  )

(define (GENMAKE-STKVERSION>=? vers)
  (unless (string? vers)
     (abort "genmake-stkversion>=? : string expected: ~s" vers))
  (unless (version>=? (version) vers)
     (abort "STk version ~a or higher expected, but your STk is version ~a.  Please obtain the newer version.\n"
            vers (version)))
  )
  

); define-module genmake

(provide "genmake")
(import genmake)  ;make exported symbol available to the current package
